/**
 * Person born or living in a country.
 *
 * @created 1993-03-15 jmn
 * @revised 1993-05-12 kw
 */

package civil_status;

import java.util.AbstractSet;
import java.util.HashSet;

abstract class Citizen {

  /** @bon Name? */
  protected /*@ spec_public non_null @*/ String name;
  /** @bon Sex? */
  //@ public invariant sex == MALE | sex == FEMALE;
  protected /*@ spec_public @*/ char sex;
  public static final char MALE = 'M', FEMALE = 'F';
  /** @bon Age? */
  //@ public invariant 0 <= age;
  protected /*@ spec_public @*/ short age;

  /**
   * @bon Spouse?
   * @bon At most one spouse allowed.
   * @design spouse reference is arity 1.
   */
  protected /*@ spec_public @*/ Citizen spouse;

  /* Close relatives, if any */ 
  /** @bon Children? */
  //@ public invariant children.elementType <: \type(Citizen);
  protected /*@ spec_public non_null @*/ AbstractSet children;
  /** @bon Parents? */
  //@ public invariant parents.elementType <: \type(Citizen);
  protected /*@ spec_public non_null @*/ AbstractSet parents;

  /**
   * @bon Single?
   * Is this citizen single?
   */
  //@ public invariant single <==> spouse == null;
  protected /*@ spec_public @*/ boolean single;

  /** @bon Each citizen has two parents. */
  /*@ public invariant parents.size() == 2; */

  /**
   * @bon May not marry children or parents or person of same sex.
   * @design Captured in specification of {@link Citizen#can_marry(Citizen)}.
   */

  /** @bon Spouse's spouse must be this person. */
  /*@ public invariant single || spouse.spouse == this; */

  /** @bon All children, if any, must have this person among their parents. */
  /*@ public invariant (\forall Citizen c; children.contains(c); 
    @                   (\exists Citizen p; parents.contains(p); p == this));
    @*/

  //@ assignable \everything;
  protected Citizen() {
    name = "Jane Doe";
    sex = FEMALE;
    age = 0;
    spouse = null;
    children = new HashSet();
    parents = new HashSet();
    single = true;
  }

  /**
   * Celebrate the wedding.
   * @param sweetheart future husband or wife
   */
  //@ requires sweetheart != null;
  //@ requires can_marry(sweetheart);
  //@ ensures spouse == sweetheart;
  abstract public void marry(Citizen sweetheart);

  /**
   * No legal hinderance?
   * @bon Impediment to marriage?
   * @param other prospective future husband or wife
   * @return true iff this citizen can marry "other"
   */
  /*@ requires other != null;
    @ ensures \result <==> (single &
    @                       other.single &
    @                       !children.contains(other) &
    @                       !parents.contains(other) &
    @                       sex != other.sex);
    @*/
  public /*@ pure @*/ boolean can_marry(Citizen other) {
    return (single &&
            other.single &&
            !children.contains(other) &&
            !parents.contains(other) &&
            sex != other.sex);
  }

  /** Admit mistake. */
  /*@ requires !single;
    @ ensures single;
    @ ensures \old(spouse).single;
    @*/
  public void divorce() {
    spouse.single = true;
    spouse.spouse = null;
    single = true;
    spouse = null;
  }
}
